import sys
from optparse import make_option

import django
from django.core.management import call_command
from django.core.management.base import BaseCommand

from chroniker.models import Job, Log

class Command(BaseCommand):
    
    help = 'Runs a specific job. The job will only run if it is not ' + \
        'currently running.'
        
    args = "job.id"
    
    option_list = getattr(BaseCommand, 'option_list', ()) + (
        make_option('--update_heartbeat',
            dest='update_heartbeat',
            default=1,
            help='If given, launches a thread to asynchronously update ' + \
                'job heartbeat status.'),
        )
        
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            parser.add_argument('args', nargs="*")
            parser.add_argument('--update_heartbeat',
                dest='update_heartbeat',
                default=1,
                help='If given, launches a thread to asynchronously update ' + \
                    'job heartbeat status.')
            self.add_arguments(parser)
        return parser
        
    def handle(self, *args, **options):
        for job_id in args:
            
            try:
                job = Job.objects.get(pk=int(job_id))
            except Job.DoesNotExist:
                sys.stderr.write("The requested Job %s does not exist.\n" % job_id)
                return
            
            # Run the job and wait for it to finish
            print('Attempting to run job %i...' % (job.id,))
            job.handle_run(update_heartbeat=int(options['update_heartbeat']))
            


from __future__ import print_function

import sys
import time
from optparse import make_option

import django
from django.core.management.base import BaseCommand

from chroniker.models import Job, Log, JobDependency

from criticalpath import Node

class Command(BaseCommand):
    help = 'Calculates the total time a series of chained jobs will take.'
    
    option_list = getattr(BaseCommand, 'option_list', ()) + (
        make_option('--samples',
            default=20,
            help='The number of log samples to use when estimating mean job run time.'),
        )
        
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            parser.add_argument('root_job_id')
            parser.add_argument('--samples',
                default=20,
                help='The number of log samples to use when estimating mean job run time.')
            self.add_arguments(parser)
        return parser
        
    def handle(self, root_job_id, **options):
        root_job = Job.objects.get(id=int(root_job_id))
        samples = int(options['samples'])
        
        # Add all system task nodes.
        system = Node('system')
        system.add(Node(root_job.id, duration=root_job.get_run_length_estimate(samples=samples)))
        print('%s takes about %s seconds' \
            % (root_job, root_job.get_run_length_estimate(samples=samples)))
        chain = root_job.get_chained_jobs()
        for job in chain:
            print('%s takes about %s seconds' \
                % (job, job.get_run_length_estimate(samples=samples)))
            node = Node(job.id, duration=job.get_run_length_estimate(samples=samples))
            node.description = job.name
            system.add(node)
        
        # Add all links between task nodes.
        print('-'*80)
        for job in chain:
            if not job.enabled:
                continue
            dependees = JobDependency.objects.filter(dependent=job, dependee__enabled=True)
            dependees = dependees.values_list('dependee_id', flat=True)
            print(job, dependees)
            for dependee in dependees:
                # Link dependent job to dependee.
                assert job.id != 1
                system.link(from_node=dependee, to_node=job.id)
        
        root_node = system.lookup_node(1)
        print('root_node:', root_node, root_node.to_nodes, root_node.incoming_nodes)
        system.add_exit()
        sys.stdout.flush()
        
        #return
        print('Updating values...')
        system.update_all()
        
        critical_path = system.get_critical_path()
        print('critical_path:', critical_path)
        system.print_times()
        
        print('min hours:', system.duration*(1/60.)*(1/60.))




from datetime import datetime, timedelta

import logging
import sys

from django.core.management.base import BaseCommand

from chroniker.models import Log

class Command(BaseCommand):
    help = 'Deletes old job logs.'
    
    def handle(self, *args, **options):
                
        if len(args) != 2:
            sys.stderr.write(
                'Command requires two arguments. '
                'Unit (weeks, days, hours or minutes) and interval.\n')
            return
        else:
            unit = str(args[0])
            if unit not in ['weeks', 'days', 'hours', 'minutes']:
                sys.stderr.write('Valid units are weeks, days, hours or minutes.\n')
                return
            try:
                amount = int(args[1]) 
            except ValueError:
                sys.stderr.write('Interval must be an integer.\n')
                return
        kwargs = {unit: amount}
        time_ago = datetime.now() - timedelta(**kwargs)
        #Log.objects.filter( run_start_datetime__lte = time_ago ).delete()
        Log.cleanup(time_ago)



from django.utils.translation import ugettext_lazy as _

YEARLY = 'YEARLY'
MONTHLY = 'MONTHLY'
WEEKLY = 'WEEKLY'
DAILY = 'DAILY'
HOURLY = 'HOURLY'
MINUTELY = 'MINUTELY'
SECONDLY = 'SECONDLY'

FREQ_CHOICES = (
    (YEARLY, _("Yearly")),
    (MONTHLY, _("Monthly")),
    (WEEKLY, _("Weekly")),
    (DAILY, _("Daily")),
    (HOURLY, _("Hourly")),
    (MINUTELY, _("Minutely")),
    (SECONDLY, _("Secondly")),
)

RRULE_WEEKDAY_DICT = {
    "MO":0,
    "TU":1,
    "WE":2,
    "TH":3,
    "FR":4,
    "SA":5,
    "SU":6,
}

DEFAULT_MONITOR_ERROR_TEMPLATE = '''
The monitor "{{ job.name }}" has indicated a problem.

Please review this monitor at {{ url }}

{{ job.monitor_description_safe }}

{{ stderr }}
'''

WALL_CLOCK_TIME = 'wall-clock-time'
CPU_TIME = 'cpu-time'
RECURSIVE_CPU_TIME = 'recursive-cpu-time'
MAX_TIME = 'max-time' # max(WALL_CLOCK_TIME, RECURSIVE_CPU_TIME)



import os
import sys

import django

PROJECT_DIR = os.path.dirname(__file__)

DATABASES = {
    'default':{
        'ENGINE': 'django.db.backends.sqlite3',
        # Don't do this. It dramatically slows down the test.
#        'NAME': '/tmp/chroniker.db',
#        'TEST_NAME': '/tmp/chroniker.db',
    }
}

ROOT_URLCONF = 'chroniker.tests.urls'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'chroniker',
    'chroniker.tests',
    'admin_steroids',
]

MEDIA_ROOT = os.path.join(PROJECT_DIR, 'media')

# Disable migrations.
# http://stackoverflow.com/a/28560805/247542
class DisableMigrations(object):

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return "notmigrations"
SOUTH_TESTS_MIGRATE = False # <= Django 1.8
# if django.VERSION > (1, 7, 0): # > Django 1.8 
#     MIGRATION_MODULES = DisableMigrations()

USE_TZ = True

AUTH_USER_MODEL = 'auth.User'

SECRET_KEY = 'abc123'

SITE_ID = 1

BASE_SECURE_URL = 'https://localhost'

BASE_URL = 'http://localhost'

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    #'django.middleware.transaction.TransactionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.locale.LocaleMiddleware',    
)

CHRONIKER_JOB_ERROR_CALLBACK = 'chroniker.tests.tests.job_error_callback'

# Required in Django>=1.10.
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            '%s/../templates' % PROJECT_DIR,
            '%s/../static' % PROJECT_DIR,
        ],
#         'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
            'loaders': [
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            ],
            'debug': True,
        },
    },
]




# Fix AttributeError caused by conflict between Django<1.7 and Python>3.
# http://stackoverflow.com/q/34827566/247542
from django.utils.six.moves import html_parser as _html_parser    
try:
    HTMLParseError = _html_parser.HTMLParseError
except AttributeError:
    # create a dummy class for Python 3.5+ where it's been removed
    class HTMLParseError(Exception):
        pass
    _html_parser.HTMLParseError = HTMLParseError



from __future__ import print_function

import sys
import importlib
from datetime import timedelta
from optparse import make_option

import django
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.utils import timezone

import six

from chroniker.models import Job, Log, get_current_job

class Command(BaseCommand):
    help = 'Runs a specific monitoring routine.'
    
    option_list = getattr(BaseCommand, 'option_list', ()) + (
        make_option('--imports',
            dest='imports',
            help='Modules to import.'),
        make_option('--query',
            dest='query',
            help='The query to run.'),
        make_option('--verbose',
            dest='verbose',
            default=False,
            help='If given, displays extra logging messages.'),
        )
        
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            parser.add_argument('args', nargs="*")
            parser.add_argument('--imports',
                dest='imports',
                help='Modules to import.')
            parser.add_argument('--query',
                dest='query',
                help='The query to run.')
            parser.add_argument('--verbose',
                dest='verbose',
                default=False,
                help='If given, displays extra logging messages.')
            self.add_arguments(parser)
        return parser
        
    def handle(self, *args, **options):
        imports = options['imports']
        query = options['query']
        verbose = options['verbose']
        assert imports, 'No imports specified.'
        assert query, 'No query specified.'
        for imp in imports.strip().split('|'):
            imp_parts = tuple(imp.split(','))
            if len(imp_parts) == 1:
                cmd = ('import %s' % imp_parts)
            elif len(imp_parts) == 2:
                cmd = ('from %s import %s' % imp_parts)
            elif len(imp_parts) == 3:
                cmd = ('from %s import %s as %s' % imp_parts)
            else:
                raise Exception('Invalid import: %s' % (imp,))
            if verbose:
                print(cmd)
            six.exec_(cmd)
        if verbose:
            print(query)
        q = eval(query, globals(), locals()) # pylint: disable=W0123
        
        job = get_current_job()
        if job:
            job.monitor_records = q.count()
            job.save()
        
        if q.count():
            print('%i records require attention.' % (q.count(),), file=sys.stderr)
        else:
            print('%i records require attention.' % (q.count(),), file=sys.stdout)




from django import forms
from django.contrib import admin
from django.contrib.admin.sites import site
from django.contrib.admin.widgets import ManyToManyRawIdWidget, ForeignKeyRawIdWidget
from django.core.urlresolvers import reverse
from django.forms.widgets import Select, TextInput, flatatt
try:
    # force_unicode was deprecated in Django 1.5.
    from django.utils.encoding import force_unicode as force_text
    from django.utils.encoding import smart_unicode as smart_text
except ImportError:
    from django.utils.encoding import force_text
    from django.utils.encoding import smart_text
from django.utils.html import escape
from django.utils.safestring import mark_safe

try:
    from django.utils.encoding import StrAndUnicode
except ImportError:
    from django.utils.encoding import python_2_unicode_compatible

    @python_2_unicode_compatible
    class StrAndUnicode:
        def __str__(self):
            return self.code

import six

from .utils import get_admin_change_url, get_admin_changelist_url

class LinkedSelect(Select):
    def render(self, name, value, attrs=None, *args, **kwargs):
        output = super(LinkedSelect, self).render(name, value, attrs=attrs, *args, **kwargs)
        model = self.choices.field.queryset.model
        to_field_name = self.choices.field.to_field_name or 'id'
        try:
            kwargs = {to_field_name:value}
            obj = model.objects.get(**kwargs)
            view_url = get_admin_change_url(obj)
            output += mark_safe('&nbsp;<a href="%s" target="_blank">view</a>&nbsp;' % (view_url,))
        except model.DoesNotExist:
            pass
        return output

class ForeignKeyTextInput(TextInput):
    """
    Implements the same markup as VerboseForeignKeyRawIdWidget but does not
    require an explicit model relationship.
    """
    
    def __init__(self, model_class, value, *args, **kwargs):
        super(ForeignKeyTextInput, self).__init__(*args, **kwargs)
        self._model_class = model_class
        self._raw_value = value
        q = model_class.objects.filter(id=value)
        self._instance = None
        if q.count():
            self._instance = q[0]
            
    def render(self, name, value, attrs=None):
        from django.template import Context, Template
        from django.template.context import Context
        if value is None:
            value = ''
        final_attrs = self.build_attrs(attrs, type=self.input_type, name=name)
        if value != '':
            # Only add the 'value' attribute if a value is non-empty.
            final_attrs['value'] = force_text(self._format_value(value))
        final_attrs['size'] = 10
        t = Template(six.u("""
{% load staticfiles %}
<input{{ attrs|safe }} />
{% if instance %}
    <a href="{{ changelist_url|safe }}?t=id" class="related-lookup" id="lookup_{{ id|safe }}" onclick="return showRelatedObjectLookupPopup(this);">
        <img src="{% static 'admin/img/selector-search.gif' %}" width="16" height="16" alt="Lookup" />
    </a>
    <strong><a href="{{ url|safe }}" target="_blank">{{ instance|safe }}</a></strong>
{% endif %}
        """))
        c = Context(dict(
            id=final_attrs['id'],
            attrs=flatatt(final_attrs),
            raw_value=self._raw_value,
            url=get_admin_change_url(self._instance),
            changelist_url=get_admin_changelist_url(self._model_class),
            instance=self._instance))
        return  mark_safe(t.render(c))

#http://djangosnippets.org/snippets/2217/

class VerboseForeignKeyRawIdWidget(ForeignKeyRawIdWidget):
    def label_for_value(self, value):
        key = self.rel.get_related_field().name
        try:
            obj = self.rel.to._default_manager.using(self.db).get(**{key: value})
            change_url = reverse(
                "admin:%s_%s_change" % (obj._meta.app_label, obj._meta.object_name.lower()),
                args=(obj.pk,)
            )
            return '&nbsp;<strong><a href="%s" target="_blank">%s</a></strong>' \
                % (change_url, escape(obj))
        except (ValueError, self.rel.to.DoesNotExist):
            return ''

class VerboseManyToManyRawIdWidget(ManyToManyRawIdWidget):
    def label_for_value(self, value):
        values = value.split(',')
        str_values = []
        key = self.rel.get_related_field().name
        for v in values:
            try:
                obj = self.rel.to._default_manager.using(self.db).get(**{key: v})
                x = smart_text(obj)
                change_url = reverse(
                    "admin:%s_%s_change" % (obj._meta.app_label, obj._meta.object_name.lower()),
                    args=(obj.pk,)
                )
                str_values += ['<strong><a href="%s" target="_blank">%s</a></strong>' \
                    % (change_url, escape(x))]
            except self.rel.to.DoesNotExist:
                str_values += ['???']
        return ', '.join(str_values)

class ImproveRawIdFieldsForm(admin.ModelAdmin):
    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name in self.raw_id_fields:
            kwargs.pop("request", None)
            typ = db_field.rel.__class__.__name__
            if typ == "ManyToOneRel":
                kwargs['widget'] = VerboseForeignKeyRawIdWidget(db_field.rel, site)
            elif typ == "ManyToManyRel":
                kwargs['widget'] = VerboseManyToManyRawIdWidget(db_field.rel, site)
            return db_field.formfield(**kwargs)
        return super(ImproveRawIdFieldsForm, self).formfield_for_dbfield(db_field, **kwargs)

class ImproveRawIdFieldsFormTabularInline(admin.TabularInline):
    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name in self.raw_id_fields:
            kwargs.pop("request", None)
            typ = db_field.rel.__class__.__name__
            if typ == "ManyToOneRel":
                kwargs['widget'] = VerboseForeignKeyRawIdWidget(db_field.rel, site)
            elif typ == "ManyToManyRel":
                kwargs['widget'] = VerboseManyToManyRawIdWidget(db_field.rel, site)
            return db_field.formfield(**kwargs)
        return super(ImproveRawIdFieldsFormTabularInline, self)\
            .formfield_for_dbfield(db_field, **kwargs)




import logging
import sys
from datetime import datetime
from time import sleep, time
from threading import Thread

from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.utils.translation import ugettext_lazy as _

from chroniker.models import Job

logger = logging.getLogger('chroniker.commands.cronserver')

class CronThread(Thread):
    daemon = True
    
    def run(self):
        logger.info("Running due jobs...")
        call_command('cron')

class Command(BaseCommand):
    args = "time"
    help = _("Emulates a reoccurring cron call to run jobs at a specified "
             "interval.  This is meant primarily for development use.")
    
    def handle(self, *args, **options):
        
        logging.basicConfig(
            stream=sys.stdout,
            level=logging.INFO,
            datefmt="%Y-%m-%d %H:%M:%S",
            format="[%(asctime)-15s] %(message)s")
        
        try:
            time_wait = int(args[0])
        except Exception:
            time_wait = 60
        
        try:
            sys.stdout.write("Starting cronserver.  Jobs will run every %d seconds.\n" % time_wait)
            sys.stdout.write("Quit the server with CONTROL-C.\n")
            
            # Wait until we're synchronized with the system clock.
            seconds = datetime.now().second
            if seconds > 0:
                sleep(60-seconds)
            
            # Run server until killed.
            while True:
                thread = CronThread()
                thread.start()
                sleep(time_wait)
        except KeyboardInterrupt:
            logger.info("Exiting...\n")
            sys.exit()



from __future__ import print_function
import sys
import time

from django.core.management.base import BaseCommand, CommandError
            
class Command(BaseCommand):
    args = ''
    help = 'A simple command that always fails.'
    
    def handle(self, *args, **options):
        print(
            'Something went wrong (but not really, this is just a test).',
            file=sys.stderr)



from django.contrib import admin
from django.contrib.auth.decorators import user_passes_test

from chroniker.admin import JobAdmin
from chroniker.models import Job

def job_run(request, pk):
    return JobAdmin(Job, admin.site).run_job_view(request, pk)
job_run = user_passes_test(lambda user: user.is_superuser)(job_run)



from __future__ import print_function
import sys
import time

import django
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    args = '[time in seconds to loop]'
    help = 'A simple command that simply sleeps for the specified duration'
    
    def create_parser(self, prog_name, subcommand):
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            parser.add_argument('target_time')
            self.add_arguments(parser)
        return parser
        
    def handle(self, target_time, **options):
        start_time = time.time()
        target_time = float(target_time)
        
        print("Sleeping for {} seconds...".format(target_time))
        time.sleep(target_time)
        
        end_time = time.time()
        print("Job ran for {} seconds".format(end_time-start_time))



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-26 22:16
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone

import chroniker

class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Job',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, verbose_name='name')),
                ('frequency', models.CharField(choices=[('YEARLY', 'Yearly'), ('MONTHLY', 'Monthly'), ('WEEKLY', 'Weekly'), ('DAILY', 'Daily'), ('HOURLY', 'Hourly'), ('MINUTELY', 'Minutely'), ('SECONDLY', 'Secondly')], max_length=10, verbose_name='frequency')),
                ('params', models.TextField(blank=True, help_text='Semicolon-separated list of <a href="http://labix.org/python-dateutil" target="_blank">rrule parameters</a>. e.g: interval:15;byhour:7,8,9', null=True, verbose_name='params')),
                ('command', models.CharField(blank=True, help_text='A valid django-admin command to run.', max_length=200, verbose_name='command')),
                ('args', models.CharField(blank=True, help_text='Space separated list; e.g: arg1 option1=True', max_length=200, verbose_name='args')),
                ('raw_command', models.CharField(blank=True, help_text='The raw shell command to run.\n            This is mutually exclusive with `command`.', max_length=1000, null=True, verbose_name='raw command')),
                ('enabled', models.BooleanField(default=True, help_text='If checked, this job will be run automatically according\n            to the frequency options.')),
                ('next_run', models.DateTimeField(blank=True, help_text="If you don't set this it will be determined automatically", null=True, verbose_name='next run')),
                ('last_run_start_timestamp', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='last run start timestamp')),
                ('last_run', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='last run end timestamp')),
                ('last_heartbeat', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='last heartbeat')),
                ('is_running', models.BooleanField(default=False)),
                ('last_run_successful', models.NullBooleanField(editable=False, verbose_name='success')),
                ('email_errors_to_subscribers', models.BooleanField(default=True, help_text='If checked, the stdout and stderr of a job will be emailed to the subscribers if an error occur.')),
                ('email_success_to_subscribers', models.BooleanField(default=False, help_text='If checked, the stdout of a job will be emailed to the subscribers if not errors occur.')),
                ('lock_file', models.CharField(blank=True, editable=False, max_length=255)),
                ('force_run', models.BooleanField(default=False, help_text='If checked, then this job will be run immediately.')),
                ('force_stop', models.BooleanField(default=False, help_text='If checked, and running then this job will be stopped.')),
                ('timeout_seconds', models.PositiveIntegerField(default=0, help_text='When non-zero, the job will be forcibly killed if\n            running for more than this amount of time.')),
                ('hostname', models.CharField(blank=True, help_text=chroniker.models.hostname_help_text_setter, max_length=700, null=True, verbose_name='target hostname')),
                ('current_hostname', models.CharField(blank=True, editable=False, help_text='The name of the host currently running the job.', max_length=700, null=True)),
                ('current_pid', models.CharField(blank=True, db_index=True, editable=False, help_text='The ID of the process currently running the job.', max_length=50, null=True)),
                ('total_parts_complete', models.PositiveIntegerField(default=0, editable=False, help_text='The total number of complete parts.')),
                ('total_parts', models.PositiveIntegerField(default=0, editable=False, help_text='The total number of parts of the task.')),
                ('is_monitor', models.BooleanField(default=False, help_text='If checked, will appear in the monitors section.')),
                ('monitor_url', models.CharField(blank=True, help_text='URL provided to further explain the monitor.', max_length=255, null=True)),
                ('monitor_error_template', models.TextField(blank=True, default='\nThe monitor "{{ job.name }}" has indicated a problem.\n\nPlease review this monitor at {{ url }}\n\n{{ job.monitor_description_safe }}\n\n{{ stderr }}\n', help_text='If this is a monitor, this is the template used to compose the error text email.<br/>Available variables: {{ job }} {{ stderr }} {{ url }}', null=True)),
                ('monitor_description', models.TextField(blank=True, help_text="An explanation of the monitor's purpose.", null=True)),
                ('monitor_records', models.IntegerField(blank=True, editable=False, help_text='The number of records that need attention.', null=True)),
                ('maximum_log_entries', models.PositiveIntegerField(default=1000, help_text='The maximum number of most recent log entries to keep.<br/>A value of 0 keeps all log entries.')),
                ('log_stdout', models.BooleanField(default=True, help_text='If checked, all characters printed to stdout will be\n            saved in a log record.')),
                ('log_stderr', models.BooleanField(default=True, help_text='If checked, all characters printed to stderr will be\n            saved in a log record.')),
                ('subscribers', models.ManyToManyField(blank=True, related_name='subscribed_jobs', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='JobDependency',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('wait_for_completion', models.BooleanField(default=True, help_text='If checked, the dependent job will not run until the dependee job has completed.')),
                ('wait_for_success', models.BooleanField(default=True, help_text='If checked, the dependent job will not run until the dependee job has completed successfully.')),
                ('wait_for_next_run', models.BooleanField(default=True, help_text='If checked, the dependent job will not run until the dependee job has a next_run greater than its next_run.')),
                ('dependee', models.ForeignKey(help_text='The thing that has other jobs waiting on it to complete.', on_delete=django.db.models.deletion.CASCADE, related_name='dependents', to='chroniker.Job')),
                ('dependent', models.ForeignKey(help_text='The thing that cannot run until another job completes.', on_delete=django.db.models.deletion.CASCADE, related_name='dependencies', to='chroniker.Job')),
            ],
            options={
                'verbose_name_plural': 'job dependencies',
            },
        ),
        migrations.CreateModel(
            name='Log',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('run_start_datetime', models.DateTimeField(db_index=True, default=django.utils.timezone.now, editable=False)),
                ('run_end_datetime', models.DateTimeField(blank=True, db_index=True, editable=False, null=True)),
                ('duration_seconds', models.PositiveIntegerField(blank=True, db_index=True, editable=False, null=True, verbose_name='duration (total seconds)')),
                ('stdout', models.TextField(blank=True)),
                ('stderr', models.TextField(blank=True)),
                ('hostname', models.CharField(blank=True, editable=False, help_text='The hostname this job was executed on.', max_length=700, null=True)),
                ('success', models.BooleanField(db_index=True, default=True, editable=False)),
                ('on_time', models.BooleanField(db_index=True, default=True, editable=False, help_text='If true, indicates job completed of its own accord.\n            If false, the job exceeded a timeout threshold and was forcibly\n            killed.')),
                ('job', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='logs', to='chroniker.Job')),
            ],
            options={
                'ordering': ('-run_start_datetime',),
            },
        ),
        migrations.CreateModel(
            name='Monitor',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('chroniker.job',),
        ),
        migrations.AlterUniqueTogether(
            name='jobdependency',
            unique_together=set([('dependent', 'dependee')]),
        ),
    ]



from __future__ import print_function
 
from setuptools import setup, find_packages, Command
from setuptools.command.test import test as TestCommand

import os
import sys
import urllib

import chroniker

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

def get_reqs(*fns):
    lst = []
    for fn in fns:
        for package in open(os.path.join(CURRENT_DIR, fn)).readlines():
            package = package.strip()
            if not package:
                continue
            lst.append(package.strip())
    return lst

setup(
    name="django-chroniker",
    version=chroniker.__version__,
    packages=find_packages(),
    scripts=['bin/chroniker'],
    package_data={
        '': ['docs/*.txt', 'docs/*.py'],
        'chroniker': [
            'static/*/*/*.*',
            'templates/*.*',
            'templates/*/*.*',
            'templates/*/*/*.*',
            'templates/*/*/*/*.*',
            'fixtures/*',
            'tests/fixtures/*',
        ],
    },
    author="Chris Spencer",
    author_email="chrisspen@gmail.com",
    description="Easily control cron jobs via Django's admin.",
    license="BSD",
    url="https://github.com/chrisspen/django-chroniker",
    #https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Framework :: Django',
    ],
    zip_safe=False,
    install_requires=get_reqs('pip-requirements-min-django.txt', 'pip-requirements.txt'),
    tests_require=get_reqs('pip-requirements-test.txt'),
)



import logging
import os
import re
import sys
import time
import errno
import socket
import subprocess
from functools import partial
from optparse import make_option
from collections import defaultdict
from multiprocessing import Process, Queue

from django.core.management.base import BaseCommand
from django.db import connection
import django
from django.conf import settings
from django.utils import timezone

import psutil

from chroniker.models import Job, Log
from chroniker import constants as c
from chroniker import utils

def kill_stalled_processes(dryrun=True):
    """
    Due to a bug in the Django|Postgres backend, occassionally
    the `manage.py cron` process will hang even through all processes
    have been marked completed.
    We compare all recorded PIDs against those still running,
    and kill any associated with complete jobs.
    """
    pids = set(map(int, Job.objects\
        .filter(is_running=False, current_pid__isnull=False)\
        .exclude(current_pid='')\
        .values_list('current_pid', flat=True)))
    for pid in pids:
        try:
            if utils.pid_exists(pid):# and not utils.get_cpu_usage(pid):
                p = psutil.Process(pid)
                cmd = ' '.join(p.cmdline())
                if 'manage.py cron' in cmd:
                    jobs = Job.objects.filter(current_pid=pid)
                    job = None
                    if jobs:
                        job = jobs[0]
                    print('Killing process %s associated with %s.' % (pid, job))
                    if not dryrun:
                        utils.kill_process(pid)
                else:
                    print('PID not cron.')
            else:
                print('PID dead.')
        except psutil.NoSuchProcess:
            print('PID does not exist.')

class JobProcess(utils.TimedProcess):
    
    def __init__(self, job, *args, **kwargs):
        super(JobProcess, self).__init__(*args, **kwargs)
        self.job = job

def run_job(job, update_heartbeat=None, **kwargs):
    
    stdout_queue = kwargs.pop('stdout_queue', None)
    stderr_queue = kwargs.pop('stderr_queue', None)
    force_run = stderr_queue('force_run', False)
    
    # TODO:causes UnicodeEncodeError: 'ascii' codec can't encode
    # character u'\xa0' in position 59: ordinal not in range(128)
    #print(u"Running Job: %i - '%s' with args: %s" \
    #    % (job.id, job, job.args))
    
    # TODO:Fix? Remove multiprocess and just running all jobs serially?
    # Multiprocessing does not play well with Django's PostgreSQL
    # connection, as it seems Django's connection code is not thread-safe.
    # It's a hacky solution, but the short-term fix seems to be to close
    # the connection in this thread, forcing Django to open a new
    # connection unique to this thread.
    # Without this call to connection.close(), we'll get the error
    # "Lost connection to MySQL server during query".
    print('Closing connection.')
    connection.close()
    print('Connection closed.')
    job.run(
        update_heartbeat=update_heartbeat,
        check_running=False,
        stdout_queue=stdout_queue,
        stderr_queue=stderr_queue,
        force_run=force_run,
    )
    #TODO:mark job as not running if still marked?
    #TODO:normalize job termination and cleanup outside of handle_run()?

def run_cron(jobs=None, update_heartbeat=True, force_run=False, dryrun=False, clear_pid=False):
    try:
        
        # TODO: auto-kill inactive long-running cron processes whose
        # threads have stalled and not exited properly?
        # Check for 0 cpu usage.
        #ps -p <pid> -o %cpu
        
        stdout_map = defaultdict(list) # {prod_id:[]}
        stderr_map = defaultdict(list) # {prod_id:[]}
        stdout_queue = Queue()
        stderr_queue = Queue()
        
        if settings.CHRONIKER_AUTO_END_STALE_JOBS and not dryrun:
            Job.objects.end_all_stale()
            
        # Check PID file to prevent conflicts with prior executions.
        # TODO: is this still necessary? deprecate? As long as jobs run by
        # JobProcess don't wait for other jobs, multiple instances of cron
        # should be able to run simeltaneously without issue.
        if settings.CHRONIKER_USE_PID:
            pid_fn = settings.CHRONIKER_PID_FN
            pid = str(os.getpid())
            any_running = Job.objects.all_running().count()
            if not any_running:
                # If no jobs are running, then even if the PID file exists,
                # it must be stale, so ignore it.
                pass
            elif os.path.isfile(pid_fn):
                try:
                    old_pid = int(open(pid_fn, 'r').read())
                    if utils.pid_exists(old_pid):
                        print('%s already exists, exiting' % pid_fn)
                        sys.exit()
                    else:
                        print(('%s already exists, but contains stale '
                            'PID, continuing') % pid_fn)
                except ValueError:
                    pass
                except TypeError:
                    pass
            open(pid_fn, 'w').write(pid)
            clear_pid = True
        
        procs = []
        if force_run:
            q = Job.objects.all()
            if jobs:
                q = q.filter(id__in=jobs)
        else:
            q = Job.objects.due_with_met_dependencies_ordered(jobs=jobs)
        
        running_ids = set()
        for job in q:
            
            # This is necessary, otherwise we get the exception
            # DatabaseError: SSL error: sslv3 alert bad record mac
            # even through we're not using SSL...
            # We work around this by forcing Django to use separate
            # connections for each process by explicitly closing the
            # current connection.
            connection.close()
            
            # Re-check dependencies to incorporate any previous iterations
            # that marked jobs as running, potentially causing dependencies
            # to become unmet.
            Job.objects.update()
            job = Job.objects.get(id=job.id)
            if not force_run and not job.is_due_with_dependencies_met(running_ids=running_ids):
                print('Job %i %s is due but has unmet dependencies.' % (job.id, job))
                continue
            
            # Immediately mark the job as running so the next jobs can
            # update their dependency check.
            print('Running job %i %s.' % (job.id, job))
            running_ids.add(job.id)
            if dryrun:
                continue
            job.is_running = True
            Job.objects.filter(id=job.id).update(is_running=job.is_running)
            
            # Launch job.
            #proc = JobProcess(job, update_heartbeat=update_heartbeat, name=str(job))
            job_func = partial(
                run_job,
                job=job,
                force_run=force_run or job.force_run,
                update_heartbeat=update_heartbeat,
                name=str(job),
            )
            proc = JobProcess(
                job=job,
                max_seconds=job.timeout_seconds,
                target=job_func,
                name=str(job),
                kwargs=dict(
                    stdout_queue=stdout_queue,
                    stderr_queue=stderr_queue,
                ))
            proc.start()
            procs.append(proc)
        
        if not dryrun:
            print("%d Jobs are due." % len(procs))
            
            # Wait for all job processes to complete.
            while procs:
                
                while not stdout_queue.empty():
                    proc_id, proc_stdout = stdout_queue.get()
                    stdout_map[proc_id].append(proc_stdout)
                    
                while not stderr_queue.empty():
                    proc_id, proc_stderr = stderr_queue.get()
                    stderr_map[proc_id].append(proc_stderr)
                    
                for proc in list(procs):
                    
                    # Auto kill processes that haven't terminated but yet
                    # register no cpu usage.
                    #cpu = proc.get_cpu_usage_recursive()
                    #print('cpu:',proc,cpu)
#                    if not cpu:
#                        utils.kill_process(proc.pid)
#                        time.sleep(1)
                    
                    if not proc.is_alive():
                        print('Process %s ended.' % (proc,))
                        procs.remove(proc)
                    elif proc.is_expired:
                        print('Process %s expired.' % (proc,))
                        proc_id = proc.pid
                        proc.terminate()
                        run_end_datetime = timezone.now()
                        procs.remove(proc)
                        
                        connection.close()
                        Job.objects.update()
                        j = Job.objects.get(id=proc.job.id)
                        run_start_datetime = j.last_run_start_timestamp
                        proc.job.is_running = False
                        proc.job.force_run = False
                        proc.job.force_stop = False
                        proc.job.save()
                        
                        # Create log record since the job was killed before it had
                        # a chance to do so.
                        Log.objects.create(
                            job=proc.job,
                            run_start_datetime=run_start_datetime,
                            run_end_datetime=run_end_datetime,
                            success=False,
                            on_time=False,
                            hostname=socket.gethostname(),
                            stdout=''.join(stdout_map[proc_id]),
                            stderr=''.join(stderr_map[proc_id]+['Job exceeded timeout\n']),
                        )
                        
                time.sleep(1)
            print('!'*80)
            print('All jobs complete!')
    finally:
        if settings.CHRONIKER_USE_PID and os.path.isfile(pid_fn) \
        and clear_pid:
            os.unlink(pid_fn)


class Command(BaseCommand):
    help = 'Runs all jobs that are due.'
    option_list = getattr(BaseCommand, 'option_list', ()) + (
        make_option('--update_heartbeat',
            dest='update_heartbeat',
            default=1,
            help='If given, launches a thread to asynchronously update ' + \
                'job heartbeat status.'),
        make_option('--force_run',
            dest='force_run',
            action='store_true',
            default=False,
            help='If given, forces all jobs to run.'),
        make_option('--dryrun',
            action='store_true',
            default=False,
            help='If given, only displays jobs to be run.'),
        make_option('--jobs',
            dest='jobs',
            default='',
            help='A comma-delimited list of job ids to limit executions to.'),
        make_option('--name',
            dest='name',
            default='',
            help='A name to give this process.'),
    )

    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            parser.add_argument('--update_heartbeat',
                dest='update_heartbeat',
                default=1,
                help='If given, launches a thread to asynchronously update ' + \
                    'job heartbeat status.')
            parser.add_argument('--force_run',
                dest='force_run',
                action='store_true',
                default=False,
                help='If given, forces all jobs to run.')
            parser.add_argument('--dryrun',
                action='store_true',
                default=False,
                help='If given, only displays jobs to be run.')
            parser.add_argument('--jobs',
                dest='jobs',
                default='',
                help='A comma-delimited list of job ids to limit executions to.')
            parser.add_argument('--name',
                dest='name',
                default='',
                help='A name to give this process.')
            self.add_arguments(parser)
        return parser
    
    def handle(self, *args, **options):
        
        kill_stalled_processes(dryrun=False)
        
        # Find specific job ids to run, if any.
        jobs = [
            int(_.strip())
            for _ in options.get('jobs', '').strip().split(',')
            if _.strip().isdigit()
        ]
        update_heartbeat = int(options['update_heartbeat'])
        force_run = options['force_run']
        run_cron(
            jobs,
            update_heartbeat=update_heartbeat,
            force_run=force_run,
            dryrun=options['dryrun'],
        )



from __future__ import print_function

import time
from optparse import make_option

import django
from django.core.management.base import BaseCommand

from chroniker.models import Job, Log

class Command(BaseCommand):
    help = 'Incrementally updates status, to help testing transaction ' + \
        'behavior on different database backends.'
    
    option_list = getattr(BaseCommand, 'option_list', ()) + (
        make_option('--seconds',
            dest='seconds',
            default=60,
            help='The number of total seconds to count up to.'),
        )
        
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            parser.add_argument('args', nargs="*")
            parser.add_argument('--seconds',
                dest='seconds',
                default=60,
                help='The number of total seconds to count up to.')
            self.add_arguments(parser)
        return parser
        
    def handle(self, *args, **options):
        seconds = int(options['seconds'])
        for i in range(seconds):
            Job.update_progress(total_parts=seconds, total_parts_complete=i)
            print('%i of %i' % (i, seconds))
            time.sleep(1)
